'' =================================================================================================
''
''   File....... jm_hc-8+_parser_demo.spin
''   Purpose.... 
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2011-2015 Jon McPhalen
''               -- see below for terms of use
''   E-mail.....  
''   Started.... 
''   Updated.... 17 MAR 2015
''
'' =================================================================================================


con { speed & timing }

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                           ' 5MHz crystal

  CLK_FREQ = (_clkmode >> 6) * _xinfreq
  MS_001   = CLK_FREQ / 1_000
  US_001   = CLK_FREQ / 1_000_000
  

con { io pins }

  RX1     = 31                                                   ' programming / debug port
  TX1     = 30
  
  SDA     = 29                                                   ' eeprom
  SCL     = 28                                                    
                                                                  
  OUT_EN  = 27                                                   ' outputs enable pin (low)     
                                                                  
  SIO     = 26                                                   ' TTL serial io
                                                                  
  RX2     = 23                                                   ' RS-485 port
  RX2_EN  = 22                                                    
  TX2_EN  = 21                                                    
  TX2     = 20                                                    
                                                                  
  R_LED   = 18                                                   ' bi-color LED
  G_LED   = 17                                                    
                                                                  
  OUT7    = 15                                                   ' outputs
  OUT6    = 14                                                    
  OUT5    = 13                                                    
  OUT4    = 12                                                    
  OUT3    = 11                                                    
  OUT2    = 10                                                    
  OUT1    =  9                                                    
  OUT0    =  8                                                    
                                                                  
  OPT_BR  =  7                                                   ' options swithes
  OPT_A1  =  6                                                    
  OPT_A0  =  5                                                    
  OPT_SM  =  4                                                    
                                                                  
  DMX_A8  =  3                                                   ' high-bit off DMX address
                                                                  
  DO_165  =  2                                                   ' control lines for 74x165
  CLK_165 =  1                                                    
  LD_165  =  0


  RX3     =  OPT_BR                                              ' external serial (e.g. Bluetooth) 
  TX3     =  OPT_A1



con { serial connection }
                                          
  #0, M_INT, M_EXT                                               ' serial connection modes

  MODE = M_INT                                                   ' use Bluetooth

  USB_BAUD = 115_200                                             ' default for terminal
  EXT_BAUD =  38_400                                             ' default for EFX-TEK BT module 


con
        
   #0, OFF, RED, GRN, YEL                                        ' for r/g led
                                                                  
   #0, NOCON, RX, TX, RXTX                                       ' for rs-485
                                                                  
   #0, LSBFIRST, MSBFIRST                                        ' for 74x165 shift-register                                               
                                                                  

obj

' main                                                           ' * primary cog (spin)

  time   : "jm_time"                                             '   timing and delays
  io     : "jm_io"                                               '   simple io
  prng   : "jm_prng"                                             '   pseudo-random number generator
  pwm    : "jm_pwm8"                                             ' * pwm driver for outputs
  fader  : "jm_fade_manager"                                     ' * controller for auto fading/cross-fading
  serial : "jm_fullduplexserial"                                 ' * serial io
  parser : "jm_parser"                                           '   string parser
  
' run_rg                                                         ' * rg led color control (spin)   
  
  
var


dat { valid characters and tokens }

  CHAR_SET              byte    "$%+-_"
                        byte    "0123456789"
                        byte    "ABCDEFGHIJKLMNOPQRSTUVWXYZ", 0   

  ' Note: For HC-8+, "pin" is virtualized output (0 = OUT0, etc.)
  
  TOKEN_LIST            { marker for address of tokens }
  
  TOKEN_HIGH            byte    "HIGH", 0                        ' HIGH pin  |  pin HIGH
  TOKEN_ON              byte    "ON", 0                          ' ON pin    |  pin ON    
  TOKEN_LOW             byte    "LOW", 0                         ' LOW pin   |  pin LOW    
  TOKEN_OFF             byte    "OFF", 0                         ' OFF pin   |  pin OFF
  TOKEN_OUTS            byte    "OUTS" , 0                       ' OUTS %bbbbbbbb

  TOKEN_BLINK           byte    "BLINK", 0                       ' BLINK pin cycles on_millis off_millis
  TOKEN_ZIPPER          byte    "ZIPPER", 0                      ' ZIPPER millis

  TOKEN_LEVEL           byte    "LEVEL", 0                       ' LEVEL pin value
  TOKEN_FADE            byte    "FADE", 0                        ' FADE ch start end millis
  TOKEN_XFADE           byte    "XFADE", 0                       ' XFADE ch1 ch2 millis

  TOKEN_RGLED           byte    "RGLED", 0                       ' RGLED color
  TOKEN_RED             byte    "RED", 0
  TOKEN_GRN             byte    "GRN", 0
  TOKEN_YEL             byte    "YEL", 0
  
  TOKEN_TTL             byte    "TTL", 0                         ' TTL (returns state of ttl pins)
  TOKEN_DMX             byte    "DMX", 0                         ' DMX (returns state of DMX address switch)

  
con { enumerated token indexes }

  ' keep list member values contiguous!

   #0, T_HIGH, T_ON, T_LOW, T_OFF, T_OUTS, {                     ' simple outputs
     } T_BLINK, T_ZIPPER,                  {                     ' advanced outputs
     } T_LEVEL, T_FADE, T_XFADE,           { 
     } T_RGLED, T_RED, T_GRN, T_YEL,       {                     ' rg led control (off valid, too)
     } T_TTL, T_DMX,                       {                     ' inputs
     
     } TOKEN_COUNT                                               ' this is the count of valid tokens
  

pub main | c, ready 

  setup_io                                                       ' configure HC-8+ io pins
  setup_objects                                                  ' configure objects used
                                                                  
  repeat
    c := serial.rxcheck                                          ' any serial input?
    if (c => 0)                                                  ' if yes, pass to parser
      ready := parser.enqueue(c)
      if (ready)                                                 ' true when tokens are available
        if (MODE == M_INT)
          report                                                 ' show command analysis
        process_cmd                                              ' process valid tokens
        parser.reset                                             ' reset parser for next command


pub report | n, tidx, p_tok

'' Display command input and tokens evaluation

  serial.str(string("--> "))
  serial.str(parser.queue_addr)                                  ' show raw command string
  serial.tx(serial#CR)
  serial.str(string("    -- length: "))                          ' characters in command
  serial.dec(parser.queue_len)
  serial.tx(serial#CR) 
  serial.str(string("    -- tokens: "))                          ' tokens in command
  serial.dec(parser.token_count)
  serial.tx(serial#CR) 

  repeat n from 0 to parser.token_count-1
    p_tok := parser.token_addr(n)                                ' get pointer to token string
    serial.str(string("    -- token"))   
    serial.dec(n)                         
    serial.str(string(": <"))                                    ' show token string in <>             
    serial.str(p_tok)                      
    serial.str(string(">, "))     
    if (parser.token_is_number(n))
      serial.str(string("number, value = "))
      serial.dec(parser.token_value(n))
    else    
      serial.str(string("string, id = "))
      tidx := parser.get_token_id(p_tok)                         ' scan tokens list for this token 
      serial.dec(tidx)
    serial.tx(serial#CR) 

  serial.tx(serial#CR)


pub process_cmd | tidx

'' Handle known tokens

  if (parser.token_is_number(0))                                 ' 1st token a number?
    process_pin
    return

  ' process string token

  tidx := parser.get_token_id(parser.token_addr(0))              ' get index of command token

  case tidx
    T_HIGH   : set_pin(true)  
    T_ON     : set_pin(true)  
    T_LOW    : set_pin(false)  
    T_OFF    : set_pin(false)
    T_OUTS   : set_outputs  
    T_BLINK  : blink_pin 
    T_ZIPPER : run_zipper
    T_LEVEL  : set_level
    T_FADE   : run_fade
    T_XFADE  : run_xfade    
    T_RGLED  : set_rgb
    T_TTL    : report_ttl
    T_DMX    : report_dmx
    other    : bad_command    


pub process_pin | pin, tidx

'' Process command that begins with #

  if (parser.token_count <> 2)                                   ' check token count
    bad_command
    return

  if (parser.token_is_number(1))                                 ' parameter must be string command (HIGH, LOW, etc)
    bad_command
    return

  pin := parser.token_value(0)                                   ' convert to number
  if ((pin < 0) or (pin > 7))                                    ' out of range?
    bad_command
    return
    
  tidx := parser.get_token_id(parser.token_addr(1))              ' get index of command token

  case tidx
    T_HIGH : pwm.high(pin)      
    T_ON   : pwm.high(pin)
    T_LOW  : pwm.low(pin)     
    T_OFF  : pwm.low(pin)


pub bad_command

{
  serial.str(string("--> "))
  serial.str(parser.queue_addr)
  serial.tx(13)
  serial.str(string("<-- Unknown/invalid command.", 13, 13))     
}


pub set_pin(state) | pin

  if (parser.token_count <> 2)                                   ' check token count
    bad_command
    return

  ' we need to use token_is_number for the pin because token_value returns 0 for non-numbers
  ' -- 0 is a valid pin  
    
  if (parser.token_is_number(1))                                 ' pin parameter must be #
    pin := parser.token_value(1)                                 ' extract value
    if ((pin < 0) or (pin > 7))                                  ' out of range?
      bad_command          
      return
  else
    bad_command
    return

  if (state)                                                     ' do it
    fader.high(pin)
  else
    fader.low(pin)


pub set_outputs | pattern

  if (parser.token_count <> 2)                                   ' check token count
    bad_command
    return

  if (parser.token_is_number(1))                                 ' pin parameter must be #
    pattern := parser.token_value(1) & $FF                       ' extract value                                                           
  else
    bad_command
    return

  fader.digital(pattern, $FF) 

    
pub blink_pin | pin, cycles, onms, offms

  if (parser.token_count <> 5)                                   ' check token count
    bad_command
    return

  ' we need to use token_is_number for the pin because token_value returns 0 for non-numbers
  ' -- 0 is a valid pin  

  if (parser.token_is_number(1))                                 ' pin parameter must be #
    pin := parser.token_value(1)                                 ' extract value
    if ((pin < 0) or (pin > 7))                                  ' out of range?          
      bad_command
      return                                                              
  else
    bad_command
    return

  cycles := parser.token_value(2)
  if (cycles < 1)
    bad_command
    return

  onms := parser.token_value(3)
  if (onms < 1)
    bad_command
    return
   
  offms := parser.token_value(4)
  if (offms < 1)
    bad_command
    return

  repeat cycles                                                  ' execute blink  
    fader.high(pin)
    time.pause(onms)
    fader.low(pin)
    time.pause(offms)

  

pub run_zipper | ms, ch

  if (parser.token_count <> 2)                                   ' check token count
    bad_command
    return

  if (parser.token_is_number(1))                                 ' ch parameter must be #
    ms := parser.token_value(1)                                  ' extract value
    if (ms < 1)
      pwm.set_all(0)    
      return
    ms <#= 1000                                                  ' cap at 1s per led                                                        
  else
    bad_command
    return

  fader.set_all(0)
    
  repeat ch from 0 to 7
    fader.high(ch)
    time.pause(ms)

  repeat ch from 0 to 7
    fader.low(ch)
    time.pause(ms)


pub set_level | ch, level

  if (parser.token_count <> 3)                                   ' check token count
    bad_command
    return

  if (parser.token_is_number(1))                                 ' ch parameter must be #
    ch := parser.token_value(1)                                  ' extract value
    if ((ch < 0) or (ch > 7))                                    ' out of range?          
      bad_command
      return                                                              
  else
    bad_command
    return

  if (parser.token_is_number(2))                                 ' pin parameter must be #
    level := parser.token_value(2)                               ' extract value
  else
    bad_command
    return

  if (MODE == M_INT)
    serial.str(string(13, "-- setting level", 13))
    

  level := fader.set(ch, level)

  {
  serial.str(string("<-- OUT"))
  serial.tx(ch + "0")
  serial.str(string(" set to "))
  serial.dec(level)
  repeat 2
    serial.tx(13)
  }  
    

pub run_fade | ch, lvlstart, lvlstop, ms

' FADE ch start end millis

  if (parser.token_count <> 5)                                   ' check token count
    bad_command
    return

  if (parser.token_is_number(1))                                 ' ch parameter must be #
    ch := parser.token_value(1)                                  ' extract value
    if ((ch < 0) or (ch > 7))                                    ' out of range?          
      bad_command
      return                                                              
  else
    bad_command
    return

  if (parser.token_is_number(2))                                 ' start parameter must be #
    lvlstart := parser.token_value(2)                            ' extract value
  else
    bad_command
    return    
    
  if (parser.token_is_number(3))                                 ' stop parameter must be #
    lvlstop := parser.token_value(3)                             ' extract value
  else
    bad_command
    return

  if (parser.token_is_number(4))                                 ' ms parameter must be #
    ms := parser.token_value(4)                                  ' extract value
    if (ms < 1)
      return
    ms <#= 10000                                                  
  else
    bad_command
    return

  fader.fade(ch, lvlstart, lvlstop, ms) 
    

pub run_xfade | ch1, ch2, ms

' XFADE ch1 ch2 millis

  if (parser.token_count <> 4)                                   ' check token count
    bad_command
    return

  if (parser.token_is_number(1))                                 ' ch1 parameter must be #
    ch1 := parser.token_value(1)                                 ' extract value
    if ((ch1 < 0) or (ch1 > 7))                                  ' out of range?          
      bad_command
      return                                                              
  else
    bad_command
    return

  if (parser.token_is_number(2))                                 ' ch1 parameter must be #
    ch2 := parser.token_value(2)                                 ' extract value
    if ((ch2 < 0) or (ch2 > 7))                                  ' out of range?          
      bad_command
      return  
    if (ch2 == ch1)                                              ' same channel
      return
  else
    bad_command
    return  
    
  if (parser.token_is_number(3))                                 ' ms parameter must be #
    ms := parser.token_value(3)                                  ' extract value
    if (ms < 1)
      return
    ms <#= 10000                                                  
  else
    bad_command
    return

  fader.cross_fade(ch1, ch2, ms)

   
 
pub set_rgb | tidx

  if (parser.token_count <> 2)                                   ' check token count
    bad_command
    return

  tidx := parser.get_token_id(parser.token_addr(1))              ' get index of command token

  case tidx
    T_OFF  : rgcolor := OFF 
    T_RED  : rgcolor := RED 
    T_GRN  : rgcolor := GRN 
    T_YEL  : rgcolor := YEL 

   
pub report_ttl 

  if (parser.token_count <> 1)                                   ' check token count
    bad_command
    return

  serial.str(string("TTL = %"))
  serial.bin(ttl_inputs(true), 8)    
  serial.tx(serial#CR)     
 

pub report_dmx | addr

  if (parser.token_count <> 1)                                   ' check token count
    bad_command
    return

  addr := dmx_address(true)
    
  serial.str(string("DMX = %"))
  serial.bin(addr, 9)
  serial.tx(" ")
  serial.tx("(")
  serial.dec(addr)
  serial.tx(")")
  serial.tx(serial#CR)    


con

  { ------------------- }
  {  P G M   S E T U P  } 
  { ------------------- }

                                                                  
pub setup_io                                                      
                                                                  
'' Configure for basic IO control                                 
                                                                  
  outa[OUT7..OUT0] := %0000_0000                                 ' clear outputs
  dira[OUT7..OUT0] := %1111_1111                                 ' output mode                                  
  io.low(OUT_EN)                                                 ' enable output driver
                                                                  
  setup_595                                                      ' setup x595 IO
                                                                  
  set_rgled(OFF)                                                  
                                                                  
  set_rs485(NOCON)                                               ' RS-485 not connected
                                                                  

pub setup_595

'' Setup IO pins for x595 shift registers
'' -- must be called from cog using the shift registers

  io.high(LD_165)                                                ' setup x165 io pins
  io.low(CLK_165)
  io.input(DO_165)
  
  
pub setup_objects

'' Setup objects used by the application

  time.start

  prng.seed(-cnt, cnt ~> 2, $EF8_4E8, cnt << 2, cnt)             ' seed the randomizer
    
  pwm.start(8, OUT0)                                             ' start outputs driver
  fader.start(8, pwm.address)                                    ' start fader controller (link to pwm object)

  if (MODE == M_INT)
    serial.start(RX1, TX1, %0000, USB_BAUD)                      ' serial via programming port
  else
    serial.start(RX3, TX3, %0000, EXT_BAUD)                      ' serial via external device (xb or bt)

  parser.start(@CHAR_SET, @TOKEN_LIST, TOKEN_COUNT, false)       ' setup parser

  cognew(rg_led, @rgstack)                                       ' start rg color cog

  
con

  { --------------------------------- }
  {  S U P P O R T   R O U T I N E S  } 
  { --------------------------------- }


var

  long  lastscan                                                 ' ttl/DMX input scan


pub ttl_pin | btns

'' Reads TTL header for 1-of-8 input
'' -- returns -1 (no input)
'' -- returns 0 to 7 for input on IN0 to IN7
''    * only one input allowed

  btns := ttl_inputs(true)                                       ' read ttl inputs
                                                                  
  if (bit_count(btns) == 1)                                      ' only one button pressed?
    return bit_pos(btns, 0)                                      '  yes, return button #
  else                                                            
    return -1                                                    '  none or bad input
                                                                  

pub ttl_inputs(rescan) 

'' Return TTL inputs status
'' -- updates global var "lastscan" if rescan is true

  if (rescan)
    lastscan := x165_in | (ina[DMX_A8] << 16)                    ' read in0..in7, dmx addresss
   
  return (lastscan & $FF)                                        ' return TTL input bits


pub dmx_address(rescan) 

'' Return DMX address switch setting
'' -- updates global var "lastscan" if rescan is true

  if (rescan)
    lastscan := x165_in | (ina[DMX_A8] << 16)                    ' read in0..in7, dmx addresss

  return (lastscan >> 8) & $01FF                                 ' return DMX address bits
 

pub x165_in | tmp165

'' Input value from dual 74x165s on board
'' -- bit order is MSBFIRST      
                                                                                          
  outa[LD_165] := 0                                              ' blip Shift/Load line    
  outa[LD_165] := 1
         
  repeat 16                                                      ' read n bits         
    tmp165 := (tmp165 << 1) | ina[DO_165]                        ' get new bit             
    outa[CLK_165] := 1                                           ' blip clock              
    outa[CLK_165] := 0
   
  return tmp165 & $FFFF
  

pub option_inputs

'' Returns state of option switch bits (%0000..%1111, all off to all on)

  dira[OPT_BR..OPT_SM] := %0000                                  ' force to input mode

  return !ina[OPT_BR..OPT_SM] & $0F                              ' return state (1 = on)


pub set_rgled(state)

'' Sets R/G LED
'' -- requires rg_led cog to be running

  case state                                                     ' set color to state
    OFF   : rgcolor := OFF                             
    RED   : rgcolor := RED                              
    GRN   : rgcolor := GRN                             
    YEL   : rgcolor := YEL
    other : rgcolor := OFF


pub set_rs485(state)

'' Sets RS-485 RX and TX enable inputs

  case state
    NOCON : outa[RX2_EN..TX2_EN] := %10
    RX    : outa[RX2_EN..TX2_EN] := %00 
    TX    : outa[RX2_EN..TX2_EN] := %11   
    RXTX  : outa[RX2_EN..TX2_EN] := %01

  dira[RX2_EN..TX2_EN] := %11


pub bit_count(value) | bc

'' Returns # of bits set (1) in value

  bc := 0                                                        ' clear count
  repeat 32                                                      ' test all bits
    bc += (value & %1)                                           ' add bit value to count
    value >>= 1                                                  ' next bit
                                                                  
  return bc                                                       


pub bit_pos(value, bitmode)

'' Returns position of 1st bit set (0..31)
'' -- mode 0 to scan from lsb, mode 1 to scan from msb 
'' -- -1 = no bits set

  if (value == 0)                                                ' if no bits                                   
    return -1                                                    '  return -1
  else                                                            
    if (bitmode == 0)                                            ' check from LSB
      value ><= 32                                               '  flip for >|
      return (32 - >|value)                                       
    else                                                          
      return (>|value - 1)


pub bit_val(value, pos)

'' Returns bit (0..1) from pos (0..31) in value 

  if ((pos => 0) and (pos =< 31))
    return (value >> pos) & 1
  else
    return 0


con

  { --------------------------- }
  {  R G L E D   C O N T R O L  } 
  { --------------------------- }


var

  long  rgcolor                                                  ' rg led state (OFF..YEL)

  long  rgstack[32]                                              ' stack space for spin cog


pri rg_led | t, rgtimer

  io.low(R_LED)                                                     ' make LED pins outputs
  io.low(G_LED)

  t := cnt
  repeat
    waitcnt(t += MS_001)
    rgtimer := ++rgtimer & %1111                                 ' keep 0 to 15    

    case rgcolor
      OFF:
        outa[R_LED..G_LED] := %00

      GRN:  
        outa[R_LED..G_LED] := %01
        
      RED:
        if (rgtimer < 6)                                         ' red = 6R:2X
          outa[R_LED..G_LED] := %10 
        else
          outa[R_LED..G_LED] := %00 

      YEL:
        if (rgtimer == 15)                                       ' yel = 1R:15G
          outa[R_LED..G_LED] := %10 
        else
          outa[R_LED..G_LED] := %01  


dat { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}      